"""
 [The "BSD licence"]
 Copyright (c) 2005-2007 Terence Parr
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from antlr4 import *


class LexerAdaptor(Lexer):

    """
      Track whether we are inside of a rule and whether it is lexical parser. _currentRuleType==Token.INVALID_TYPE
      means that we are outside of a rule. At the first sign of a rule name reference and _currentRuleType==invalid, we
      can assume that we are starting a parser rule. Similarly, seeing a token reference when not already in rule means
      starting a token rule. The terminating ';' of a rule, flips this back to invalid type.

      This is not perfect logic but works. For example, "grammar T;" means that we start and stop a lexical rule for
      the "T;". Dangerous but works.

      The whole point of this state information is to distinguish between [..arg actions..] and [charsets]. Char sets
      can only occur in lexical rules and arg actions cannot occur.
    """

    _currentRuleType = Token.INVALID_TYPE

    def __init__(self, inp, output):
        Lexer.__init__(self, inp, output)

    def getCurrentRuleType(self):
        return self._currentRuleType

    def setCurrentRuleType(self, ruleType):
        self._currentRuleType = ruleType

    def handleBeginArgument(self):
        if self.inLexerRule():
            self.pushMode(self.LexerCharSet)
            self.more()
        else:
            self.pushMode(self.Argument)

    def handleEndArgument(self):
        self.popMode()
        if len(self._modeStack) > 0:
            self._type = self.ARGUMENT_CONTENT

    def handleEndAction(self):
        self.popMode()
        if len(self._modeStack) > 0:
            self._type = self.ACTION_CONTENT

    def emit(self):
        if self._type == self.ID:
            firstChar = self._input.getText(self._tokenStartCharIndex, self._tokenStartCharIndex)
            if firstChar[0].isupper():
                self._type = self.TOKEN_REF
            else:
                self._type = self.RULE_REF

            if self._currentRuleType == Token.INVALID_TYPE:  # if outside of rule def
                self._currentRuleType = self._type  # set to inside lexer or parser rule

        elif self._type == self.SEMI:  # exit rule def
            self._currentRuleType = Token.INVALID_TYPE
        return Lexer.emit(self)

    def inLexerRule(self):
        return self._currentRuleType == self.TOKEN_REF

    def inParserRule(self):  # not used, but added for clarity
        return self._currentRuleType == self.RULE_REF



# Copyright (c) 2017-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from .hdd_tree import HDDRule, HDDToken


def remove_empty_nodes(node):
    """
    Delete those nodes from the HDD tree that don't contribute to the output at
    all (tokens with empty text, e.g., the EOF token; and rules with no
    children, e.g., lambda rules).

    :param node: The root of the tree to be transformed.
    :return: The root of the transformed tree.
    """
    if isinstance(node, HDDRule):
        non_empty_children = []

        for child in node.children:
            if isinstance(child, HDDToken):
                # empty token is usually the EOF only (but interestingly, it may
                # appear multiple times in the tree)
                if child.text != '':
                    non_empty_children.append(child)
            else:
                assert isinstance(child, HDDRule)
                remove_empty_nodes(child)

                # a grammar may contain lambda rules (with nothing on the
                # right-hand side, or with an empty alternative), or rules that
                # produce EOF only (which is removed in the branch above)
                if child.children:
                    non_empty_children.append(child)

        node.children[:] = non_empty_children

    return node


def flatten_recursion(node):
    """
    Heuristics to flatten left or right-recursion. E.g., given a rule
        rule : a | rule b
    and a HDD tree built with it from an input, rewrite the resulting HDD tree
    as if it was built using
        rule : a b*
    This allows HDD to potentially completely remove the recurring blocks
    (instead of replacing them with their minimal replacement, which is usually
    not "").

    :param node: The root of the tree to be transformed.
    :return: The root of the transformed tree.
    """
    if isinstance(node, HDDRule) and node.state == node.KEEP:
        for child in node.children:
            flatten_recursion(child)

        if len(node.children) > 1 and node.name:
            if node.children[0].name == node.name:
                left = node.children[0]

                right = HDDRule('', replace='', start=node.children[1].start, end=node.children[-1].end)
                right.add_children(node.children[1:])
                del node.children[:]

                node.add_children(left.children)
                node.add_child(right)

            elif node.children[-1].name == node.name:
                right = node.children[-1]

                left = HDDRule('', replace='', start=node.children[0].start, end=node.children[-2].end)
                left.add_children(node.children[0:-1])
                del node.children[:]

                node.add_child(left)
                node.add_children(right.children)

        # This only seems to happen if there was some error during parsing.
        # In this case a weird 1-step chain gets inserted into the left/right-
        # recursive tree, which prevents flattening. But we cannot postpone the
        # merging of this 1-step chain to squeeze_tree because flatten_recursion
        # is usually not called again afterwards. So, do a degenerate "rotation"
        # (i.e., simple lifting) here.
        if len(node.children) == 1 and node.name:
            if node.children[0].name == node.name:
                child = node.children[0]
                del node.children[:]
                node.add_children(child.children)

    return node


def squeeze_tree(node):
    """
    Compress single line chains in the HDD tree whose minimal replacements are
    the same and hence they would result in redundant checks during the
    minimization.

    :param node: The root of the tree to be transformed.
    :return: The root of the transformed tree.
    """
    if isinstance(node, HDDRule):
        for i, child in enumerate(node.children):
            squeezed_child = squeeze_tree(child)
            if child != squeezed_child:
                node.children[i].replace_with(squeezed_child)

        if len(node.children) == 1 and node.children[0].replace == node.replace:
            return node.children[0]

    return node


def skip_unremovable(node, *, unparse_with_whitespace=True):
    """
    Mark those nodes as removed whose unparsing (e.g., for tokens, their text)
    is the same tokens as their minimal replacement, thus hiding them from
    hddmin, because they just cause extra test runs but cannot reduce the input.

    :param node: The root of the tree to be transformed.
    :return: The root of the transformed tree.
    """
    if isinstance(node, HDDRule):
        for child in node.children:
            skip_unremovable(child, unparse_with_whitespace=unparse_with_whitespace)

    if node.unparse(with_whitespace=unparse_with_whitespace) == node.replace:
        node.state = node.REMOVED

    return node


def skip_whitespace(node):
    """
    Mark tokens with whitespace-only text as removed. Useful when hidden-channel
    tokens are built into the tree to let hddmin deal with
    hidden-but-non-whitespace tokens only.

    :param node: The root of the tree to be transformed.
    :return: The root of the transformed tree.
    """
    if isinstance(node, HDDRule):
        for child in node.children:
            skip_whitespace(child)
    else:
        assert isinstance(node, HDDToken)
        if node.text.isspace():
            node.state = node.REMOVED

    return node



# Copyright (c) 2018-2021 Renata Hodovan, Akos Kiss.
# Copyright (c) 2021 Daniel Vince
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import itertools
import logging

from .prune import prune

logger = logging.getLogger(__name__)


def hddrmin(hdd_tree, *,
            reduce_class, reduce_config, tester_class, tester_config,
            id_prefix=(), cache=None, unparse_with_whitespace=True,
            config_filter=None, transformations=(prune,), hdd_star=True,
            pop_first=False, append_reversed=False):
    """
    Run the recursive variant of the hierarchical delta debugging reduce
    algorithm (a.k.a. HDDr).

    The tree traversal implementation is actually not recursive but an iterative
    queue-based reformulation of HDDr. How tree nodes are popped from the queue
    during the iteration (whether from the beginning or from the end of the
    queue) and how the children of a visited node are appended to the queue
    (whether they are added in forward or reverse order) give rise to different
    variants of HDDr:

        - 'pop first' with 'forward append' gives the classic breadth-first
          traversal,
        - 'pop first' with 'reverse append' gives syntactically reversed
          breadth-first traversal,
        - 'pop last' with 'reverse append' gives the classic depth-first
          traversal,
        - 'pop last' with 'forward append' gives syntactically reversed
          depth-first traversal.

    :param hdd_tree: The root of the tree that the reduce will work with (it's
        the output of create_hdd_tree).
    :param reduce_class: Reference to the reducer class (DD, ParallelDD or
        CombinedParallelDD from the picire module).
    :param reduce_config: Dictionary containing the parameters of the
        reduce_class init function.
    :param tester_class: Reference to a callable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing the parameters of the tester
        class init function (except test_builder).
    :param id_prefix: Tuple to prepend to config IDs during tests.
    :param cache: Cache to use.
    :param unparse_with_whitespace: Build test case by adding whitespace between
        nonadjacent tree nodes during unparsing.
    :param config_filter: Filter function from node to boolean, to allow running
        hddmin selectively.
    :param transformations: Iterable of transformations that reduce a
        configuration of nodes.
    :param hdd_star: Boolean to enable the HDD star algorithm.
    :param pop_first: Boolean to control tree traversal (see above for details).
    :param append_reverse: Boolean to control tree traversal (see above for
        details).
    :return: The reduced test case (1-tree-minimal if hdd_star is True and
        config_filter is None).
    """

    for iter_cnt in itertools.count():
        logger.info('Iteration #%d', iter_cnt)

        changed = False
        queue = [hdd_tree]
        for node_cnt in itertools.count():
            if not queue:
                break
            if pop_first:
                queue, node = queue[1:], queue[0]
            else:
                queue, node = queue[:-1], queue[-1]
            if not hasattr(node, 'children') or node.state != node.KEEP:
                continue

            children = [child for child in node.children if child.state == child.KEEP]
            if config_filter:
                children = list(filter(config_filter, children))

            if children:
                logger.info('Checking node #%d ...', node_cnt)

                for trans_cnt, transformation in enumerate(transformations):
                    hdd_tree, transformed = transformation(hdd_tree, children,
                                                           reduce_class=reduce_class, reduce_config=reduce_config,
                                                           tester_class=tester_class, tester_config=tester_config,
                                                           id_prefix=id_prefix + ('i%d' % iter_cnt, 'n%d' % node_cnt, 't%d' % trans_cnt),
                                                           cache=cache,
                                                           unparse_with_whitespace=unparse_with_whitespace)

                    changed = changed or transformed

            for child in node.children if not append_reversed else reversed(node.children):
                if child.state == child.KEEP:
                    queue.append(child)

        if not hdd_star or not changed:
            break

    return hdd_tree



# Copyright (c) 2007 Ghassan Misherghi.
# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
# Copyright (c) 2021 Daniel Vince
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import itertools
import logging

from .info import height
from .prune import prune

logger = logging.getLogger(__name__)


def hddmin(hdd_tree, *,
           reduce_class, reduce_config, tester_class, tester_config,
           id_prefix=(), cache=None, unparse_with_whitespace=True,
           config_filter=None, transformations=(prune,), hdd_star=True):
    """
    Run the hierarchical delta debugging reduce algorithm.

    :param hdd_tree: The root of the tree that the reduce will work with (it's
        the output of create_hdd_tree).
    :param reduce_class: Reference to the reducer class (DD, ParallelDD or
        CombinedParallelDD from the picire module).
    :param reduce_config: Dictionary containing the parameters of the
        reduce_class init function.
    :param tester_class: Reference to a callable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing the parameters of the tester
        class init function (except test_builder).
    :param id_prefix: Tuple to prepend to config IDs during tests.
    :param cache: Cache to use.
    :param unparse_with_whitespace: Build test case by adding whitespace between
        nonadjacent tree nodes during unparsing.
    :param config_filter: Filter function from node to boolean, to allow running
        hddmin selectively.
    :param transformations: Iterable of transformations that reduce a
        configuration of nodes.
    :param hdd_star: Boolean to enable the HDD star algorithm.
    :return: The reduced test case (1-tree-minimal if hdd_star is True and
        config_filter is None).
    """

    def collect_level_nodes(level):
        def _collect_level_nodes(node, current_level):
            if node.state != node.KEEP:
                return
            if current_level == level:
                level_nodes.append(node)
            elif hasattr(node, 'children'):
                for child in node.children:
                    _collect_level_nodes(child, current_level + 1)
        level_nodes = []  # Using `list` (not `set`) for the sake of stability.
        _collect_level_nodes(hdd_tree, 0)
        return level_nodes

    for iter_cnt in itertools.count():
        logger.info('Iteration #%d', iter_cnt)

        changed = False
        for level in itertools.count():
            level_nodes = collect_level_nodes(level)
            if not level_nodes:
                break

            if config_filter:
                level_nodes = list(filter(config_filter, level_nodes))
                if not level_nodes:
                    continue

            logger.info('Checking level %d / %d ...', level, height(hdd_tree))

            for trans_cnt, transformation in enumerate(transformations):
                hdd_tree, transformed = transformation(hdd_tree, level_nodes,
                                                       reduce_class=reduce_class, reduce_config=reduce_config,
                                                       tester_class=tester_class, tester_config=tester_config,
                                                       id_prefix=id_prefix + ('i%d' % iter_cnt, 'l%d' % level, 't%d' % trans_cnt),
                                                       cache=cache,
                                                       unparse_with_whitespace=unparse_with_whitespace)

                changed = changed or transformed

        if not hdd_star or not changed:
            break

    return hdd_tree



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from os import listdir
from os.path import basename, commonprefix, split, splitext
from subprocess import CalledProcessError, PIPE, run, STDOUT

logger = logging.getLogger(__name__)
grammar_cache = {}


def build_grammars(grammars, out, antlr, lang='python'):
    """
    Build lexer and grammar from ANTLRv4 grammar files in Python target.

    :param grammars: Tuple of grammar files.
    :param out: Output directory.
    :param antlr: Path to the ANTLR4 tool (Java jar binary).
    :param lang: The target language of the parser.
    :return: List of references/names of the lexer, parser and listener classes
        of the target.
    """

    # Generate parser and lexer in the target language and return either with
    # python class ref or the name of java classes.
    if lang not in grammar_cache:
        grammar_cache[lang] = {}
    if grammars in grammar_cache[lang]:
        logger.debug('%r is already built with %s target.', grammars, lang)
        return grammar_cache[lang][grammars]

    try:
        languages = {
            'python': {'antlr_arg': '-Dlanguage=Python3', 'ext': 'py', 'listener_format': 'Listener'},
            'java': {'antlr_arg': '-Dlanguage=Java', 'ext': 'java', 'listener_format': 'BaseListener'},
        }

        try:
            run(('java', '-jar', antlr, languages[lang]['antlr_arg'], '-o', out) + grammars,
                stdout=PIPE, stderr=STDOUT, cwd=out, check=True)
        except CalledProcessError as e:
            logger.error('Building grammars %r failed!\n%s\n', grammars, e.output)
            raise

        files = listdir(out)
        filename = basename(grammars[0])

        def file_endswith(end_pattern):
            f = next(f for f in files if len(commonprefix([filename, f])) > 0 and f.endswith(end_pattern))
            _, f = split(f)
            f, _ = splitext(f)
            return f

        # Extract the name of lexer and parser from their path.
        lexer = file_endswith('Lexer.{ext}'.format(ext=languages[lang]['ext']))
        parser = file_endswith('Parser.{ext}'.format(ext=languages[lang]['ext']))
        # The name of the generated listeners differs if Python or other language target is used.
        listener = file_endswith('{listener_format}.{ext}'.format(listener_format=languages[lang]['listener_format'], ext=languages[lang]['ext']))

        if lang == 'python':
            grammar_cache[lang][grammars] = [getattr(__import__(x, globals(), locals(), [x], 0), x) for x in [lexer, parser, listener]]
        else:
            grammar_cache[lang][grammars] = [lexer, parser, listener]

        return grammar_cache[lang][grammars]
    except Exception as e:
        logger.error('Exception while loading parser modules', exc_info=e)
        raise



# Copyright (c) 2018-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import xml.etree.ElementTree as ET

from subprocess import CalledProcessError, PIPE, run

from ..hdd_tree import HDDRule, HDDToken, Position
from ..transform import remove_empty_nodes


logger = logging.getLogger(__name__)


def build_hdd_tree(element, start):
    name = element.tag
    name = name.replace('{http://www.srcML.org/srcML/src}', 'src:')
    name = name.replace('{http://www.srcML.org/srcML/cpp}', 'cpp:')
    name = name.replace('{http://www.srcML.org/srcML/position}', 'pos:')

    rule = HDDRule(name, start=start, end=start, replace='')
    result = [rule]

    if element.text:
        end = start.after(element.text)
        rule.add_child(HDDToken('{name}@text'.format(name=name), element.text, start=start, end=end, replace=element.text))
        rule.end = end

    for child in list(element):
        if child.tag.startswith('{http://www.srcML.org/srcML/position}'):
            continue
        for node in build_hdd_tree(child, rule.end):
            rule.add_child(node)
            rule.end = rule.children[-1].end

    if element.tail:
        result += [HDDToken('{name}@tail'.format(name=name), element.tail, start=rule.end, end=rule.end.after(element.tail), replace=element.tail)]

    return result


def create_hdd_tree(src, *, language):
    """
    Build a tree that the HDD algorithm can work with.

    :param src: Input source to srcML.
    :param language: Language of the input source (C, C++, C#, or Java).
    :return: The root of the created HDD tree.
    """

    try:
        stdout = run(('srcml', '--language={language}'.format(language=language)),
                     input=src, stdout=PIPE, stderr=PIPE, check=True).stdout
    except CalledProcessError as e:
        logger.error('Parsing with srcml failed!\n%s\n%s\n', e.stdout, e.stderr)
        raise

    root = ET.fromstring(stdout)

    tree_result = build_hdd_tree(root, Position())
    assert len(tree_result) == 1
    tree = tree_result[0]

    tree = remove_empty_nodes(tree)
    return tree



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import json
import logging
import re
import shutil
import sys

from glob import glob
from os import makedirs, pathsep
from os.path import basename, join
from pkgutil import get_data
from string import Template
from subprocess import CalledProcessError, PIPE, run, STDOUT

from antlr4 import CommonTokenStream, error, InputStream, Token
from antlr4.Token import CommonToken

from .grammar_analyzer import analyze_grammars
from .parser_builder import build_grammars
from ..hdd_tree import HDDRule, HDDToken, Position
from ..transform import remove_empty_nodes


logger = logging.getLogger(__name__)


class HDDQuantifier(HDDRule):
    """
    Special rule type in the HDD tree to support optional quantifiers.
    """
    def __init__(self, *, start=None, end=None):
        super().__init__('', start=start, end=end)


class HDDHiddenToken(HDDToken):
    """
    Special token type that represents tokens from hidden channels.
    """
    pass


class HDDErrorToken(HDDToken):
    """
    Special token type that represents unmatched tokens. The minimal replacement
    of such nodes is an empty string.
    """
    def __init__(self, text, *, start=None, end=None):
        super().__init__('', text, start=start, end=end)


# Override ConsoleErrorListener to suppress parse issues in non-verbose mode.
class ConsoleListener(error.ErrorListener.ConsoleErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        logger.debug('line %d:%d %s', line, column, msg)


error.ErrorListener.ConsoleErrorListener.INSTANCE = ConsoleListener()


def create_hdd_tree(src, *,
                    input_format, start,
                    antlr, lang='python',
                    hidden_tokens=False,
                    work_dir):
    """
    Build a tree that the HDD algorithm can work with.

    :param src: Input source.
    :param input_format: Dictionary describing the input format.
    :param start: Name of the start rule in [grammarname:]rulename format.
    :param antlr: Path to the ANTLR4 tool (Java jar binary).
    :param lang: The target language of the parser.
    :param hidden_tokens: Build hidden tokens of the input format into the HDD
        tree.
    :param work_dir: Working directory.
    :return: The root of the created HDD tree.
    """

    def inject_optional_actions(grammar, positions, target_file):
        """
        Update the original parser grammar by injecting actions to the start and
        end of every quantified part.

        :param grammar: Path to the grammar to be updated.
        :param positions: Start and end locations of quantified elements.
        :param target_file: Path to the updated grammar.
        """
        with open(grammar, 'rb') as f:
            lines = f.read().splitlines(True)

        languages = {
            'python': {
                'prefix': b'({self.enter_optional()} ',
                'postfix': b' {self.exit_optional()})'
            },
            'java': {
                'prefix': b'({ try { getClass().getMethod("enter_optional").invoke(this); } catch (Exception e) { assert false; }} ',
                'postfix': b' { try { getClass().getMethod("exit_optional").invoke(this); } catch (Exception e) { assert false; }})'
            }
        }

        for ln in positions:
            offset = 0
            for position in sorted(positions[ln], key=lambda x: x[1]):
                if position[0] == 's':
                    lines[ln - 1] = lines[ln - 1][0:position[1] + offset] + languages[lang]['prefix'] + lines[ln - 1][position[1] + offset:]
                    offset += len(languages[lang]['prefix'])
                elif position[0] == 'e':
                    lines[ln - 1] = lines[ln - 1][0:position[1] + offset] + languages[lang]['postfix'] + lines[ln - 1][position[1] + offset:]
                    offset += len(languages[lang]['postfix'])

        with open(target_file, 'wb') as f:
            f.write(b''.join(lines))

    def java_classpath(current_workdir):
        return pathsep.join([antlr, current_workdir])

    def compile_java_sources(lexer, parser, listener, current_workdir):
        executor = Template(get_data(__package__, join('resources', 'ExtendedTargetParser.java')).decode('utf-8'))
        with open(join(current_workdir, 'Extended{parser}.java'.format(parser=parser)), 'w') as f:
            f.write(executor.substitute(dict(lexer_class=lexer,
                                             parser_class=parser,
                                             listener_class=listener)))
        try:
            run(('javac', '-classpath', java_classpath(current_workdir)) + tuple(basename(j) for j in glob(join(current_workdir, '*.java'))),
                stdout=PIPE, stderr=STDOUT, cwd=current_workdir, check=True)
        except CalledProcessError as e:
            logger.error('Java compile failed!\n%s\n', e.output)
            raise

    def prepare_parsing(grammar_name):
        """
        Performs initiative steps needed to parse the input test case (like
        create directory structures, builds grammars, sets PATH, etc...)

        :param grammar_name: Name of the grammar to use for parsing.
        """
        grammar = input_format[grammar_name]
        resources = [fn for fn in grammar['files'] if not fn.endswith('.g4')]
        grammar['files'] = [fn for fn in grammar['files'] if fn.endswith('.g4')]

        replacements, action_positions = analyze_grammars(grammar['files'], grammar['replacements'])
        logger.debug('Replacements are calculated...')

        current_workdir = join(work_dir, grammar_name) if grammar_name else work_dir
        makedirs(current_workdir, exist_ok=True)
        if current_workdir not in sys.path:
            sys.path.append(current_workdir)

        # Inject actions into the target grammars to help localizing part of the test case that are optional.
        for i, g in enumerate(grammar['files']):
            grammar['files'][i] = join(current_workdir, basename(g))
            inject_optional_actions(g, action_positions[g], grammar['files'][i])

        for r in resources:
            shutil.copy(r, current_workdir)

        target_lexer_class, target_parser_class, target_listener_class = build_grammars(tuple(grammar['files']), current_workdir, antlr, lang)
        logger.debug('Target grammars are processed...')

        if lang == 'java':
            compile_java_sources(target_lexer_class, target_parser_class, target_listener_class, current_workdir)
            input_format[grammar_name].update(lexer=target_lexer_class, parser=target_parser_class, listener=target_listener_class, replacements=replacements)
            return

        class ExtendedTargetLexer(target_lexer_class):
            """
            ExtendedTargetLexer is a subclass of the original lexer
            implementation. It can recognize skipped tokens and instead of
            eliminating them from the parser they can be redirected to the
            dedicated PICIRENY_CHANNEL for later use.
            """

            PICIRENY_CHANNEL = -3

            # Skipped tokens cannot be accessed from the parser but we still need them to
            # unparse test cases correctly. Sending these tokens to a dedicated channel won't
            # alter the parse but makes these tokens available.
            def skip(self):
                self._channel = self.PICIRENY_CHANNEL

        class ExtendedTargetParser(target_parser_class):
            """
            ExtendedTargetParser is a subclass of the original parser
            implementation. It can trigger state changes that are needed to
            identify parts of the input that are not needed to keep it
            syntactically correct.
            """
            def enter_optional(self):
                self.trigger_listener('enter_optional')

            def exit_optional(self):
                self.trigger_listener('exit_optional')

            def enterRecursionRule(self, localctx, state, ruleIndex, precedence):
                super().enterRecursionRule(localctx, state, ruleIndex, precedence)
                self.trigger_listener('recursion_enter')

            def pushNewRecursionContext(self, localctx, state, ruleIndex):
                super().pushNewRecursionContext(localctx, state, ruleIndex)
                self.trigger_listener('recursion_push')

            def unrollRecursionContexts(self, parentCtx):
                super().unrollRecursionContexts(parentCtx)
                self.trigger_listener('recursion_unroll')

            def trigger_listener(self, event):
                for listener in self.getParseListeners():
                    if hasattr(listener, event):
                        getattr(listener, event)()

            def syntax_error_warning(self):
                if self.getNumberOfSyntaxErrors() > 0:
                    logger.warning('%s finished with %d syntax errors. This may decrease reduce quality.',
                                   target_parser_class.__name__, self.getNumberOfSyntaxErrors())

        class ExtendedTargetListener(target_listener_class):
            """
            ExtendedTargetListener is a subclass of the original listener
            implementation. It can trigger state changes that are needed to
            identify parts of the input that are not needed to keep it
            syntactically correct.
            """
            def __init__(self, parser):
                self.parser = parser
                self.current_node = None
                self.root = None
                self.seen_terminal = False
                self.island_nodes = []

            def recursion_enter(self):
                assert isinstance(self.current_node, HDDRule)
                node = HDDRule(self.current_node.name)
                self.current_node.add_child(node)
                self.current_node.recursive_rule = True
                self.current_node = node

            def recursion_push(self):
                assert self.current_node.parent.children

                first_child = self.current_node.parent.children[0]
                self.current_node.parent.remove_child(first_child)
                self.current_node.add_child(first_child)

            def recursion_unroll(self):
                assert self.current_node.recursive_rule
                assert len(self.current_node.children) == 1 and self.current_node.name == self.current_node.children[0].name
                children_to_lift = self.current_node.children[0].children
                parent = self.current_node.parent
                if children_to_lift:
                    self.current_node.children = []
                    self.current_node.add_children(children_to_lift)
                else:
                    parent.remove_child(self.current_node)
                self.current_node = parent

            def enterEveryRule(self, ctx):
                name = self.parser.ruleNames[ctx.getRuleIndex()]
                node = HDDRule(name)
                if not self.root:
                    self.root = node
                else:
                    assert self.current_node
                    self.current_node.add_child(node)
                self.current_node = node

            def exitEveryRule(self, ctx):
                # If the input contains syntax error, then the last optional block was may not closed.
                while isinstance(self.current_node, HDDQuantifier):
                    self.exit_optional()

                assert self.current_node.name == self.parser.ruleNames[ctx.getRuleIndex()], \
                    '%s (%s) != %s' % (self.current_node.name, repr(self.current_node), self.parser.ruleNames[ctx.getRuleIndex()])

                if self.current_node.parent:
                    self.current_node = self.current_node.parent

            def tokenBoundaries(self, token):
                start = Position(token.line, token.column)
                return start, start.after(token.text)

            def addToken(self, node, child):
                if not self.seen_terminal:
                    hidden_tokens = self.parser.getTokenStream().getHiddenTokensToLeft(node.symbol.tokenIndex, -1) or []
                    for token in hidden_tokens:
                        start, end = self.tokenBoundaries(token)
                        self.current_node.add_child(HDDHiddenToken(self.parser.symbolicNames[token.type], token.text,
                                                                   start=start, end=end))
                self.seen_terminal = True

                self.current_node.add_child(child)

                hidden_tokens = self.parser.getTokenStream().getHiddenTokensToRight(node.symbol.tokenIndex, -1) or []
                for token in hidden_tokens:
                    start, end = self.tokenBoundaries(token)
                    self.current_node.add_child(HDDHiddenToken(self.parser.symbolicNames[token.type], token.text,
                                                               start=start, end=end))

            def visitTerminal(self, node):
                token = node.symbol
                name, text = (self.parser.symbolicNames[token.type], token.text) if token.type != Token.EOF else ('EOF', '')
                start, end = self.tokenBoundaries(token)

                child = HDDToken(name, text, start=start, end=end)
                self.addToken(node, child)
                if name in grammar['islands']:
                    self.island_nodes.append(child)

            def visitErrorNode(self, node):
                if hasattr(node, 'symbol'):
                    token = node.symbol
                    start, end = self.tokenBoundaries(token)
                    self.addToken(node, HDDErrorToken(token.text, start=start, end=end))

            def enter_optional(self):
                quant_node = HDDQuantifier()
                self.current_node.add_child(quant_node)
                self.current_node = quant_node

            def exit_optional(self):
                assert self.current_node.parent, 'Quantifier node has no parent.'
                assert self.current_node.children, 'Quantifier node has no children.'

                self.current_node = self.current_node.parent

        input_format[grammar_name].update(lexer=ExtendedTargetLexer, parser=ExtendedTargetParser, listener=ExtendedTargetListener, replacements=replacements)

    class ExtendedErrorListener(error.ErrorListener.ErrorListener):

        def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
            t = CommonToken(source=(recognizer, recognizer._input),
                            type=Token.INVALID_TYPE,
                            channel=Token.DEFAULT_CHANNEL,
                            start=recognizer._tokenStartCharIndex,
                            stop=recognizer._tokenStartCharIndex)
            t.line = recognizer._tokenStartLine
            t.column = recognizer._tokenStartColumn
            recognizer._type = Token.MIN_USER_TOKEN_TYPE
            recognizer.emitToken(t)

    def build_hdd_tree(src, grammar_name, start_rule):
        """
        Parse the input with the provided ANTLR classes.

        :param src: Input source.
        :param grammar_name: Name of the grammar to use for parsing.
        :param start_rule: The name of the start rule of the parser.
        :return: The root of the created HDD tree.
        """

        grammar = input_format[grammar_name]
        island_nodes = []

        def set_replacement(node):
            if isinstance(node, (HDDQuantifier, HDDErrorToken)):
                node.replace = ''
            elif isinstance(node, HDDRule):
                node.replace = grammar['replacements'][node.name]
            else:
                node.replace = grammar['replacements'].get(node.name, node.text)

            if isinstance(node, HDDRule):
                for child in node.children:
                    set_replacement(child)

        logger.debug('Parse input with %s rule', start_rule)
        if lang != 'python':

            def hdd_tree_from_json(node_dict):
                # Convert interval dictionaries to Position objects.
                if 'start' in node_dict:
                    node_dict['start'] = Position(**node_dict['start'])
                if 'end' in node_dict:
                    node_dict['end'] = Position(**node_dict['end'])

                name = node_dict.get('name', None)
                children = node_dict.pop('children', None)
                cls = globals()[node_dict.pop('type')]
                node = cls(**node_dict)

                if children:
                    for child in children:
                        node.add_child(hdd_tree_from_json(child))
                elif name:
                    if name in grammar['islands']:
                        island_nodes.append(node)
                return node

            try:
                current_workdir = join(work_dir, grammar_name) if grammar_name else work_dir
                proc = run(('java', '-classpath', java_classpath(current_workdir), 'Extended{parser}'.format(parser=grammar['parser']), start_rule),
                           input=src, stdout=PIPE, stderr=PIPE, universal_newlines=True, cwd=current_workdir, check=True)
                if proc.stderr:
                    logger.debug(proc.stderr)
                result = json.loads(proc.stdout)
                tree_root = hdd_tree_from_json(result)
            except CalledProcessError as e:
                logger.error('Java parser failed!\n%s\n%s', e.stdout, e.stderr)
                raise
        else:
            lexer = grammar['lexer'](InputStream(src))
            lexer.addErrorListener(ExtendedErrorListener())
            target_parser = grammar['parser'](CommonTokenStream(lexer))
            parser_listener = grammar['listener'](target_parser)
            target_parser.addParseListener(parser_listener)

            getattr(target_parser, start_rule)()
            target_parser.syntax_error_warning()
            island_nodes = parser_listener.island_nodes
            assert parser_listener.root == parser_listener.current_node
            tree_root = parser_listener.root

        # Traverse the HDD tree and set minimal replacements for nodes.
        set_replacement(tree_root)
        process_island_nodes(island_nodes, grammar['islands'])
        logger.debug('Parse done.')
        return tree_root

    def process_island_nodes(island_nodes, island_format):
        for node in island_nodes:
            if not isinstance(island_format[node.name], tuple):
                rewritten, mapping = rename_regex_groups(island_format[node.name])
                for new_name, old_name in mapping.items():
                    grammar_name, rule_name = split_grammar_rule_name(old_name)
                    mapping[new_name] = (grammar_name, rule_name)
                    if 'lexer' not in input_format[grammar_name]:
                        prepare_parsing(grammar_name)
                island_format[node.name] = (re.compile(rewritten, re.S), mapping)

            new_node = HDDRule(node.name, replace=node.replace)
            new_node.add_children(build_island_subtree(node, *island_format[node.name]))
            node.replace_with(new_node)

    def build_island_subtree(node, pattern, mapping):
        """
        Process terminal with an island grammar.

        :param node: HDDToken object containing island language.
        :return: List of HDDTree nodes representing the `children` of node.
        """
        last_processed = 0
        content = node.text
        children = []

        # Intervals describes a non-overlapping splitting of the content according to the pattern.
        intervals = []
        for m in re.finditer(pattern, content):
            intervals.extend((g, m.start(g), m.end(g)) for g in list(pattern.groupindex.keys()) if m.start(g) != m.end(g))
        intervals.sort(key=lambda x: (x[1], x[2]))

        def shift_positions(node, start):
            if node.start:
                node.start.shift(start)
            if node.end:
                node.end.shift(start)

            if isinstance(node, HDDRule):
                for child in node.children:
                    shift_positions(child, start)

        for interval in intervals:
            # Create simple HDDToken of the substring proceeding a subgroup.
            if last_processed < interval[1]:
                token_start = node.start.after(content[0:last_processed])
                token_text = content[last_processed:interval[1]]
                children.append(HDDToken('', token_text,
                                         start=token_start,
                                         end=token_start.after(token_text),
                                         replace=token_text))

            # Process an island and save its subtree.
            island_start = node.start.after(content[0:interval[1]])
            island_root = build_hdd_tree(src=content[interval[1]:interval[2]],
                                         grammar_name=mapping[interval[0]][0],
                                         start_rule=mapping[interval[0]][1])
            shift_positions(island_root, island_start)
            children.append(island_root)

            last_processed = interval[2]

        # Create simple HDDToken of the substring following the last subgroup if any.
        if last_processed < len(content):
            token_start = node.start.after(content[0:last_processed])
            token_text = content[last_processed:]
            children.append(HDDToken('', token_text,
                                     start=token_start,
                                     end=token_start.after(token_text),
                                     replace=token_text))
        return children

    def calculate_rule_boundaries(node):
        if isinstance(node, HDDRule):
            for child in node.children:
                calculate_rule_boundaries(child)

            node.start = node.children[0].start
            node.end = node.children[-1].end

        return node

    def remove_hidden_tokens(node):
        if isinstance(node, HDDRule):
            non_hidden_children = []

            for child in node.children:
                if not isinstance(child, HDDHiddenToken):
                    remove_hidden_tokens(child)
                    non_hidden_children.append(child)

            node.children[:] = non_hidden_children

        return node

    _NAMED_GRP_PATTERN = re.compile(r'(?<!\\)(\(\?P<[^>]*>)')   # "(?P<NAME>" not prefixed by a "\"
    _NAMED_GRP_PREFIX = '(?P<'
    _NAMED_GRP_SUFFIX = '>'
    _NAMED_REF_PATTERN = re.compile(r'(?<!\\)(\(\?P=[^)]*\))')  # "(?P=NAME)" not prefixed by a "\"
    _NAMED_REF_PREFIX = '(?P='
    _NAMED_REF_SUFFIX = ')'

    def rename_regex_groups(pattern):
        """
        Rewrite capture group names in a regex pattern to ensure that the names
        are valid Python identifiers (as expected by the ``re`` module). This
        enables more sophisticated capture group names than allowed by default.

        :param str pattern: the original regex pattern with potentially extended
            syntax for capture group names.
        :return: the rewritten regex pattern and a mapping from the newly
            introduced capture group names (which are guaranteed to by valid
            Python identifiers) to the names used in the original pattern.
        :rtype: tuple(str, dict(str, str))

        .. note::

           The function expects ``pattern`` to be syntactically valid. Its
           behavior is undefined for erroneous input.
        """

        grp_rewritten = ''
        mapping = {}
        rmapping = {}
        cnt = 1
        for item in _NAMED_GRP_PATTERN.split(pattern):
            if _NAMED_GRP_PATTERN.match(item):
                old_name = item[len(_NAMED_GRP_PREFIX):-len(_NAMED_GRP_SUFFIX)]
                new_name = 'G' + str(cnt)
                cnt += 1

                mapping[new_name] = old_name
                rmapping[old_name] = new_name

                item = _NAMED_GRP_PREFIX + new_name + _NAMED_GRP_SUFFIX

            grp_rewritten += item

        ref_rewritten = ''
        for item in _NAMED_REF_PATTERN.split(grp_rewritten):
            if _NAMED_REF_PATTERN.match(item):
                old_name = item[len(_NAMED_REF_PREFIX):-len(_NAMED_REF_SUFFIX)]
                new_name = rmapping.get(old_name, old_name)

                item = _NAMED_REF_PREFIX + new_name + _NAMED_REF_SUFFIX

            ref_rewritten += item

        return ref_rewritten, mapping

    def split_grammar_rule_name(name):
        """
        Determine the grammar and the rule parts in a potentially
        grammar-prefixed rule name. The syntax for the prefixed format is
        "[grammar:]rule", where "[]" denote optionality and the default for a
        missing grammar part is the empty string.

        :param str name: a potentially grammar-prefixed rule name.
        :return: a 2-tuple of the grammar and the rule name parts.
        :rtype: tuple(str, str)
        """

        names = name.split(':', 1)
        if len(names) < 2:
            names.insert(0, '')
        return names[0], names[1]

    start_grammar, start_rule = split_grammar_rule_name(start)
    prepare_parsing(start_grammar)
    tree = build_hdd_tree(src=src,
                          grammar_name=start_grammar,
                          start_rule=start_rule)
    if not hidden_tokens:
        tree = remove_hidden_tokens(tree)
    tree = remove_empty_nodes(tree)
    tree = calculate_rule_boundaries(tree)
    return tree



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from . import cli
from . import info
from . import transform
from .cli import __version__, build_with_antlr4, build_with_srcml, reduce
from .hdd import hddmin
from .hddr import hddrmin
from .hdd_tree import HDDRule, HDDToken, HDDTree



# Copyright (c) 2007 Ghassan Misherghi.
# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from itertools import count
from os import linesep


class Position(object):
    """
    Class defining a position in the input file. Used to recognise line breaks
    between tokens.
    """
    def __init__(self, line=1, column=0):
        """
        Initialize position object.

        :param line: Line number in the input (starts with 1).
        :param column: Character index relative to the beginning of the line
            (starts with 0).

        Note: The numbering of lines (1-based) and columns (0-based) follows
        ANTLR v4.
        """
        self.line = line
        self.column = column

    def after(self, text):
        """
        Calculate the end position of a text starting at the current position.
        """
        line_breaks = text.count('\n')
        return Position(self.line + line_breaks,
                        self.column + len(text) if not line_breaks else len(text) - text.rfind('\n') - 1)

    def shift(self, start):
        """
        Shift the position by prepending a starting position.
        """
        if self.line > 1:
            self.line += start.line - 1
        else:
            self.line = start.line
            self.column += start.column

    def __repr__(self):
        return '%s(%r, %r)' % (self.__class__.__name__, self.line, self.column)


class HDDTree(object):
    # Node states for unparsing.
    REMOVED = 0
    KEEP = 1

    # ID generator
    __id = count()

    def __init__(self, name, *, start=None, end=None, replace=None):
        """
        Initialize a HDD tree/node.

        :param name: The name of the node.
        :param start: Position object describing the start of the HDDTree node.
        :param end: Position object describing the end of the HDDTree node.
        :param replace: The minimal replacement string of the current node.
        """
        self.name = name
        self.replace = replace
        self.start = start
        self.end = end
        self.parent = None
        self.state = self.KEEP
        self.id = next(self.__id)

    def unparse(self, *, with_whitespace=True, transform=None):
        """
        Build test case from a HDD tree.

        :param with_whitespace: Add whitespace (space, new line) to separate
            nonadjacent nodes.
        :param transform: A function applied to each node before unparsing, or
            None.
        :return: The unparsed test case.
        """
        def _unparse(node):
            if transform:
                node = transform(node)

            if node.state != node.KEEP:
                return node.replace

            # Keep the text of the token.
            if isinstance(node, HDDToken):
                return node.text

            if not node.children:
                return ''

            # Concat the text of children.
            child_strs = [_unparse(child) for child in node.children]
            node_str = child_strs[0]
            for i in range(1, len(node.children)):
                # Do not add extra spaces if the next chunk is empty.
                if not child_strs[i]:
                    continue
                if with_whitespace:
                    if node.children[i].start.line > node.children[i - 1].end.line:
                        node_str += linesep
                    elif node.children[i].start.column > node.children[i - 1].end.column:
                        node_str += ' '
                node_str += child_strs[i]

            return node_str

        return _unparse(self)

    def replace_with(self, other):
        """
        Replace the current node with `other` in the HDD tree.

        :param other: Node to replace the current with.
        """
        self.parent.children[self.parent.children.index(self)] = other
        other.parent = self.parent


class HDDToken(HDDTree):
    def __init__(self, name, text, *, start=None, end=None, replace=None):
        super().__init__(name, start=start, end=end, replace=replace)
        self.text = text

    def __repr__(self):
        parts = [
            'name=%r' % self.name,
            'text=%r' % self.text,
        ]
        if self.replace is not None:
            parts.append('replace=%r' % self.replace)
        if self.start is not None:
            parts.append('start=%r' % self.start)
        if self.end is not None:
            parts.append('end=%r' % self.end)
        parts.append('id=%r' % self.id)
        if self.state != self.KEEP:
            parts.append('state=%r' % self.state)

        return '%s(%s)' % (self.__class__.__name__, ', '.join(parts))


class HDDRule(HDDTree):
    def __init__(self, name, *, start=None, end=None, replace=None):
        super().__init__(name, start=start, end=end, replace=replace)
        self.children = []

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def add_children(self, children):
        for child in children:
            self.add_child(child)

    def remove_child(self, child):
        self.children.remove(child)

    def __repr__(self):
        def _indent(text, prefix):
            return ''.join(prefix + line for line in text.splitlines(True))

        parts = [
            'name=%r' % self.name,
        ]
        if self.replace is not None:
            parts.append('replace=%r' % self.replace)
        if self.start is not None:
            parts.append('start=%r' % self.start)
        if self.end is not None:
            parts.append('end=%r' % self.end)
        parts.append('id=%r' % self.id)
        if self.state != self.KEEP:
            parts.append('state=%r' % self.state)
        if self.state == self.KEEP and self.children:
            parts.append('children=[\n%s\n]' % _indent(',\n'.join(repr(child) for child in self.children), '  '))

        return '%s(%s)' % (self.__class__.__name__, ', '.join(parts))



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import os
import pytest
import subprocess
import sys


is_windows = sys.platform.startswith('win32')
script_ext = '.bat' if is_windows else '.sh'

tests_dir = os.path.dirname(os.path.abspath(__file__))
resources_dir = os.path.join(tests_dir, 'resources')
antlr = os.getenv('ANTLR')


@pytest.mark.parametrize('test, inp, exp, grammar, rule, input_format', [
    ('test-json-obj-arr-foo', 'inp-obj-arr.json', 'exp-obj-arr-foo.json', 'JSON.g4', 'json', None),
    ('test-json-obj-arr-bar', 'inp-obj-arr.json', 'exp-obj-arr-bar.json', 'JSON.g4', 'json', None),
    ('test-json-obj-arr-baz', 'inp-obj-arr.json', 'exp-obj-arr-baz.json', 'JSON.g4', 'json', None),
    ('test-json-obj-arr-87', 'inp-obj-arr.json', 'exp-obj-arr-87.json', 'JSON.g4', 'json', None),
    ('test-inijson-str-arr-87', 'inp-str-arr.ini', 'exp-str-arr-87.ini', None, None, 'inijson-crlf.json' if is_windows else 'inijson.json'),
])
@pytest.mark.parametrize('args', [
    ('--cache=config', ),
    ('--no-skip-unremovable', '--parser=java', '--cache=content', ),
    ('--no-squeeze-tree', '--parser=java', '--cache=none', ),
    ('--no-squeeze-tree', '--no-skip-unremovable', '--cache=config', ),
    ('--no-hdd-star', '--parser=java', '--cache=content', ),
    ('--no-hdd-star', '--no-skip-unremovable', '--cache=none', ),
    ('--no-hdd-star', '--no-squeeze-tree', '--cache=config', ),
    ('--no-hdd-star', '--no-squeeze-tree', '--no-skip-unremovable', '--parser=java', '--cache=content', ),
    ('--parallel', ),
])
def test_cli(test, inp, exp, grammar, rule, input_format, args, tmpdir):
    out_dir = str(tmpdir)
    cmd = (sys.executable, '-m', 'picireny') \
          + ('--test=' + test + script_ext, '--input=' + inp, '--out=' + out_dir) \
          + ('--log-level=TRACE', )
    if grammar:
        cmd += ('--grammar=' + grammar, )
    if rule:
        cmd += ('--start=' + rule, )
    if input_format:
        cmd += ('--format=' + input_format, )
    if antlr:
        cmd += ('--antlr=' + antlr, )
    cmd += args
    subprocess.run(cmd, cwd=resources_dir, check=True)

    with open(os.path.join(out_dir, inp), 'rb') as outf:
        outb = outf.read()
    with open(os.path.join(resources_dir, exp), 'rb') as expf:
        expb = expf.read()
    assert outb == expb



# Copyright (c) 2016-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import json

from argparse import ArgumentParser
from os.path import abspath, dirname, exists, join, realpath
from shutil import rmtree

try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

import antlerinator
import inators
import picire

from inators import log as logging

from . import filter, hdd, hddr, hoist, info, prune, transform

logger = logging.getLogger('picireny')
__version__ = metadata.version(__package__)


args_hdd_choices = {
    'hdd': hdd.hddmin,
    'hddr': hddr.hddrmin,
}


args_phase_choices = {
    'prune': {'transformations': [prune.prune]},
    'coarse-prune': {'transformations': [prune.prune], 'config_filter': filter.coarse_filter},
    'hoist': {'transformations': [hoist.hoist]},
    'prune+hoist': {'transformations': [prune.prune, hoist.hoist]},
    'coarse-prune+hoist': {'transformations': [prune.prune, hoist.hoist], 'config_filter': filter.coarse_filter}
}


def process_antlr4_args(args):
    antlerinator.process_antlr_argument(args)
    args.antlr = realpath(args.antlr)

    def load_format_config(data):
        # Interpret relative grammar paths compared to the directory of the config file.
        if 'files' in data:
            for i, fn in enumerate(data['files']):
                path = join(abspath(dirname(args.format)), fn)
                if not exists(path):
                    raise ValueError('Invalid input format definition: %s, defined in the format config, does not exist.' % path)
                data['files'][i] = path
            data['islands'] = data.get('islands', {})
            data['replacements'] = data.get('replacements', {})
        return data

    args.input_format = {}

    if args.format:
        if not exists(args.format):
            raise ValueError('Invalid input format definition: %s does not exist.' % args.format)

        with open(args.format, 'r') as f:
            try:
                input_description = json.load(f, object_hook=load_format_config)
                args.input_format = input_description['grammars']
                if not args.start:
                    args.start = input_description.get('start', None)
            except ValueError as e:
                raise ValueError('Invalid input format definition: The content of %s is not a valid JSON object.' % args.format) from e

    if not args.start:
        raise ValueError('Invalid input format definition: No start has been defined.')

    if args.grammar or args.replacements:
        # Initialize the default grammar that doesn't need to be named.
        if '' not in args.input_format:
            args.input_format[''] = {'files': [], 'replacements': {}, 'islands': {}}

        if args.grammar:
            for i, g in enumerate(args.grammar):
                args.input_format['']['files'].append(realpath(g))
                if not exists(args.input_format['']['files'][i]):
                    raise ValueError('Invalid input format definition: %s does not exist.' % args.input_format['']['files'][i])

        if args.replacements:
            if not exists(args.replacements):
                raise ValueError('Invalid input format definition: %s does not exist.' % args.replacements)

            try:
                with open(args.replacements, 'r') as f:
                    args.input_format['']['replacements'] = json.load(f)
            except ValueError as e:
                raise ValueError('Invalid input format definition: The content of %s is not a valid JSON object.' % args.replacements) from e


def process_srcml_args(args):
    if not args.srcml_language:
        raise ValueError('The following argument is required for srcML: --srcml:language')


def process_args(args):
    inators.arg.process_log_level_argument(args, logger)
    inators.arg.process_sys_recursion_limit_argument(args)

    args.hddmin = args_hdd_choices[args.hdd]
    args.hdd_phase_configs = [args_phase_choices[phase] for phase in (args.phase or ['prune'])]

    if args.builder == 'antlr4':
        process_antlr4_args(args)
    elif args.builder == 'srcml':
        process_srcml_args(args)

    picire.cli.process_args(args)


def log_tree(title, hdd_tree):
    logger.debug('%s\n\theight: %s\n\tshape: %s\n\tnodes: %s\n',
                 title,
                 info.height(hdd_tree),
                 ', '.join('%s' % cnt for cnt in info.shape(hdd_tree)),
                 ', '.join('%d %s' % (cnt, ty) for ty, cnt in sorted(info.count(hdd_tree).items())))
    logger.trace('%r', hdd_tree)


def build_with_antlr4(src, *,
                      input_format, start,
                      antlr, lang='python',
                      build_hidden_tokens=False,
                      work_dir):
    """
    Execute ANTLRv4-based tree building part of picireny as if invoked from
    command line, however, control its behaviour not via command line arguments
    but function parameters.

    :param src: Contents of the test case to reduce.
    :param input_format: Dictionary describing the input format.
    :param start: Name of the start rule in [grammarname:]rulename format.
    :param antlr: Path to the ANTLR4 tool (Java jar binary).
    :param lang: The target language of the parser.
    :param build_hidden_tokens: Build hidden tokens of the input format into the
        HDD tree.
    :param work_dir: Path to a working directory.
    :return: The built HDD tree.
    """
    # Get the parameters in a dictionary so that they can be pretty-printed
    args = locals().copy()
    del args['src']
    picire.cli.log_args('Building tree with ANTLRv4', args)

    from .antlr4 import create_hdd_tree
    return create_hdd_tree(src,
                           input_format=input_format, start=start,
                           antlr=antlr, lang=lang,
                           hidden_tokens=build_hidden_tokens,
                           work_dir=work_dir)


def build_with_srcml(src, *, language):
    """
    Execute srcML-based tree building part of picireny as if invoked from
    command line, however, control its behaviour not via command line arguments
    but function parameters.

    :param src: Contents of the test case to reduce.
    :param language: Language of the input source (C, C++, C#, or Java).
    :return: The built HDD tree.
    """
    # Get the parameters in a dictionary so that they can be pretty-printed
    args = locals().copy()
    del args['src']
    picire.cli.log_args('Building tree with srcML', args)

    from .srcml import create_hdd_tree
    return create_hdd_tree(src, language=language)


def reduce(hdd_tree, *,
           hddmin, reduce_class, reduce_config, tester_class, tester_config,
           cache_class=None, unparse_with_whitespace=True,
           hdd_phase_configs=({},), hdd_star=True,
           flatten_recursion=False, squeeze_tree=True, skip_unremovable=True, skip_whitespace=False):
    """
    Execute tree reduction part of picireny as if invoked from command line,
    however, control its behaviour not via command line arguments but function
    parameters.

    :param hdd_tree: HDD tree to reduce.
    :param hddmin: Function implementing a HDD minimization algorithm.
    :param reduce_class: Reference to the reducer class.
    :param reduce_config: Dictionary containing information to initialize the
        reduce_class.
    :param tester_class: Reference to a runnable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing information to initialize the
        tester_class.
    :param cache_class: Reference to the cache class to use.
    :param unparse_with_whitespace: Unparse by adding whitespace between
        nonadjacent nodes.
    :param hdd_phase_configs: Sequence of dictionaries containing information to
        parametrize the hddmin function.
    :param hdd_star: Boolean to enable the HDD star algorithm.
    :param flatten_recursion: Boolean to enable flattening left/right-recursive
        trees.
    :param squeeze_tree: Boolean to enable the tree squeezing optimization.
    :param skip_unremovable: Boolean to enable hiding unremovable nodes from
        ddmin.
    :param skip_whitespace: Boolean to enable hiding whitespace-only tokens from
        ddmin.
    :return: The reduced HDD tree.
    """
    # Get the parameters in a dictionary so that they can be pretty-printed
    args = locals().copy()
    del args['hdd_tree']
    picire.cli.log_args('Reduce session starts', args)

    log_tree('Initial tree', hdd_tree)

    # Perform tree transformations.
    if flatten_recursion:
        hdd_tree = transform.flatten_recursion(hdd_tree)
        log_tree('Tree after recursion flattening', hdd_tree)

    if squeeze_tree:
        hdd_tree = transform.squeeze_tree(hdd_tree)
        log_tree('Tree after squeezing', hdd_tree)

    if skip_unremovable:
        hdd_tree = transform.skip_unremovable(hdd_tree, unparse_with_whitespace=unparse_with_whitespace)
        log_tree('Tree after skipping unremovable nodes', hdd_tree)

    if skip_whitespace:
        hdd_tree = transform.skip_whitespace(hdd_tree)
        log_tree('Tree after skipping whitespace tokens', hdd_tree)

    # Perform reduction.
    for phase_cnt, phase_config in enumerate(hdd_phase_configs):
        logger.info('Phase #%d', phase_cnt)
        hdd_tree = hddmin(hdd_tree,
                          reduce_class=reduce_class, reduce_config=reduce_config,
                          tester_class=tester_class, tester_config=tester_config,
                          id_prefix=('p%d' % phase_cnt,),
                          cache=cache_class() if cache_class else None,
                          unparse_with_whitespace=unparse_with_whitespace,
                          hdd_star=hdd_star,
                          **phase_config)
        log_tree('Tree after reduction phase #%d' % phase_cnt, hdd_tree)

    return hdd_tree


def execute():
    """
    The main entry point of picireny.
    """
    logging.basicConfig(format='%(message)s')

    arg_parser = ArgumentParser(description='CLI for the Picireny Hierarchical Delta Debugging Framework',
                                parents=[picire.cli.create_parser()], add_help=False)

    # General HDD settings.
    arg_parser.add_argument('--builder', metavar='NAME', choices=['antlr4', 'srcml'], default='antlr4',
                            help='tool to build tree representation from input (%(choices)s; default: %(default)s)')
    arg_parser.add_argument('--hdd', metavar='NAME', choices=args_hdd_choices.keys(), default='hdd',
                            help='HDD variant to run (%(choices)s; default: %(default)s)')
    arg_parser.add_argument('--phase', metavar='NAME', choices=args_phase_choices.keys(), action='append',
                            help='parametrization of the HDD variant to run (%(choices)s; default: prune) '
                                 '(may be specified multiple times to run different parametrizations in sequence)')
    arg_parser.add_argument('--no-hdd-star', dest='hdd_star', default=True, action='store_false',
                            help='run the hddmin algorithm only once')
    arg_parser.add_argument('--flatten-recursion', default=False, action='store_true',
                            help='flatten recurring blocks of left/right-recursive rules')
    arg_parser.add_argument('--no-squeeze-tree', dest='squeeze_tree', default=True, action='store_false',
                            help='don\'t squeeze rule chains in tree representation')
    arg_parser.add_argument('--no-skip-unremovable', dest='skip_unremovable', default=True, action='store_false',
                            help='don\'t hide unremovable nodes from the ddmin algorithm')
    arg_parser.add_argument('--skip-whitespace', dest='skip_whitespace', default=False, action='store_true',
                            help='hide whitespace tokens from the ddmin algorithm')
    inators.arg.add_sys_recursion_limit_argument(arg_parser)
    inators.arg.add_version_argument(arg_parser, version=__version__)

    # ANTLRv4-specific settings.
    antlr4_grp = arg_parser.add_argument_group('ANTLRv4-specific arguments')
    antlr4_grp.add_argument('-s', '--start', '--antlr4:start', metavar='NAME',
                            help='name of the start rule in [grammarname:]rulename format (default for '
                                 'the optional grammarname is the empty string)')
    antlr4_grp.add_argument('-g', '--grammar', '--antlr4:grammar', metavar='FILE', nargs='+',
                            help='grammar file(s) describing the input format (these grammars will be '
                                 'associated with the empty grammar name, see `--start`)')
    antlr4_grp.add_argument('-r', '--replacements', '--antlr4:replacements', metavar='FILE',
                            help='JSON file defining the default replacements for lexer and parser '
                                 'rules of the grammar with the empty name (usually defined via `--grammar`)')
    antlr4_grp.add_argument('--format', '--antlr4:format', metavar='FILE',
                            help='JSON file describing a (possibly complex) input format')
    antlr4_grp.add_argument('--build-hidden-tokens', '--antlr4:build-hidden-tokens', default=False, action='store_true',
                            help='build hidden tokens of the grammar(s) into the HDD tree')
    antlerinator.add_antlr_argument(antlr4_grp, long_alias='--antlr4:antlr')
    antlr4_grp.add_argument('--parser', '--antlr4:parser', metavar='LANG', default='python', choices=['python', 'java'],
                            help='language of the generated parsers (%(choices)s; default: %(default)s) '
                                 '(using Java might gain performance, but needs JDK)')

    # srcML-specific settings.
    srcml_grp = arg_parser.add_argument_group('srcML-specific arguments')
    srcml_grp.add_argument('--srcml:language', dest='srcml_language', metavar='LANG', choices=['C', 'C++', 'C#', 'Java'],
                           help='language of the input (%(choices)s; default: %(default)s)')

    args = arg_parser.parse_args()

    try:
        process_args(args)
    except ValueError as e:
        arg_parser.error(e)

    if args.builder == 'antlr4':
        work_dir = join(args.out, 'grammar')
        hdd_tree = build_with_antlr4(args.src,
                                     input_format=args.input_format, start=args.start,
                                     antlr=args.antlr, lang=args.parser,
                                     build_hidden_tokens=args.build_hidden_tokens,
                                     work_dir=work_dir)
        unparse_with_whitespace = not args.build_hidden_tokens
        if args.cleanup:
            rmtree(work_dir)
    elif args.builder == 'srcml':
        hdd_tree = build_with_srcml(args.src, language=args.srcml_language)
        unparse_with_whitespace = False

    hdd_tree = reduce(hdd_tree,
                      hddmin=args.hddmin,
                      reduce_class=args.reduce_class, reduce_config=args.reduce_config,
                      tester_class=args.tester_class, tester_config=args.tester_config,
                      cache_class=args.cache, unparse_with_whitespace=unparse_with_whitespace,
                      hdd_phase_configs=args.hdd_phase_configs, hdd_star=args.hdd_star,
                      flatten_recursion=args.flatten_recursion,
                      squeeze_tree=args.squeeze_tree,
                      skip_unremovable=args.skip_unremovable,
                      skip_whitespace=args.skip_whitespace)
    out_src = hdd_tree.unparse(with_whitespace=unparse_with_whitespace)

    picire.cli.postprocess(args, out_src)



# Copyright (c) 2018-2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from .hdd_tree import HDDRule


def count(node, *, removed=False):
    """
    Count nodes in the tree by type.

    Note: If `removed` is `True`, removed tokens and rules are also counted (but
    sub-trees of removed rules are not).

    :param node: The root of the tree to do the counting for.
    :return: A dictionary of counts indexed by node type name.
    """
    def _count(node):
        if node.state != node.KEEP and not removed:
            return

        ty = node.__class__.__name__
        if ty not in stats:
            stats[ty] = 0
        stats[ty] += 1

        if isinstance(node, HDDRule) and node.state == node.KEEP:
            for child in node.children:
                _count(child)

    stats = {}
    _count(node)
    return stats


def height(node, *, removed=False):
    """
    Calculate the height of the tree.

    Note: If `removed` is `True`, removed tokens and rules are also counted (but
    sub-trees of removed rules are not).

    :param node: The root of the tree to do the calculation for.
    :return: The height of the tree.
    """
    if node.state != node.KEEP and not removed:
        return 0

    return 1 + (max((height(child) for child in node.children), default=0)
                if isinstance(node, HDDRule) and node.state == node.KEEP else 0)


def shape(node, *, removed=False):
    """
    Calculate the shape of the tree, i.e., the number of nodes on each tree
    level.

    Note: If `removed` is `True`, removed tokens and rules are also counted (but
    sub-trees of removed rules are not).

    :param node: The root of the tree to do the calculation for.
    :return: A list of level sizes.
    """
    def _shape(node, level):
        if node.state != node.KEEP and not removed:
            return

        if len(sizes) <= level:
            sizes.extend([0] * (level - len(sizes) + 1))
        sizes[level] += 1

        if isinstance(node, HDDRule) and node.state == node.KEEP:
            for child in node.children:
                _shape(child, level + 1)

    sizes = []
    _shape(node, 0)
    return sizes



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from copy import copy

from picire import AbstractDD, Outcome

logger = logging.getLogger(__name__)


class PruningTestBuilder(object):

    def __init__(self, tree, ids, *, with_whitespace=True):
        """
        Initialize the test builder.

        :param tree: Tree representing the current test case.
        :param ids: The IDs of nodes that can change status.
        :param with_whitespace: Unparse by adding whitespace between nonadjacent
            nodes.
        """
        self.tree = tree
        self.ids = ids
        self.with_whitespace = with_whitespace

    def __call__(self, config):
        """
        :param config: List of IDs of nodes that will be kept in the next test
            case.
        :return: The unparsed test case containing only the units defined in
            config.
        """
        def removed(node):
            if node.id in self.ids and node.id not in config:
                removed_node = copy(node)
                removed_node.state = removed_node.REMOVED
                return removed_node
            return node

        config = set(config)
        return self.tree.unparse(with_whitespace=self.with_whitespace, transform=removed)


class EmptyDD(AbstractDD):
    """
    Special DD variant that *does* test the empty configuration (and nothing
    else).
    """

    def __init__(self, test, *, cache=None, id_prefix=()):
        """
        Initialize an EmptyDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        """
        super().__init__(test=test, split=None, cache=cache, id_prefix=id_prefix)

    def __call__(self, config):
        """
        Return a 1-minimal failing subset of the initial configuration, and also
        test the empty configuration while doing so.

        Note: The initial configuration is expected to be of size 1, thus the
        1-minimal failing subset is always its trivial subset: either itself or
        the empty configuration.

        :param config: The initial configuration that will be reduced.
        :return: 1-minimal failing configuration.
        """
        assert len(config) == 1
        # assert self._test_config(config, ('assert',)) == self.FAIL

        empty = []
        config_id = ('empty',)

        logger.info('Run #empty')
        logger.info('\tConfig size: %d', len(config))
        logger.debug('\tConfig: %r', config)

        outcome = self._lookup_cache(empty, config_id) or self._test_config(empty, config_id)
        if outcome == Outcome.FAIL:
            config = empty
            logger.info('\tReduced')

        logger.info('\tDone')
        return config


def prune(hdd_tree, config_nodes, *,
          reduce_class, reduce_config, tester_class, tester_config,
          id_prefix, cache, unparse_with_whitespace):
    """
    Pruning-based reduction of a set of nodes (i.e., sub-trees), as used by
    various hierarchical delta debugging algorithm variants.

    :param hdd_tree: The root of the tree.
    :param config_nodes: The list of nodes to reduce.
    :param reduce_class: Reference to the reducer class (DD, ParallelDD or
        CombinedParallelDD from the picire module).
    :param reduce_config: Dictionary containing the parameters of the
        reduce_class init function.
    :param tester_class: Reference to a callable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing the parameters of the tester
        class init function (except test_builder).
    :param id_prefix: Tuple to prepend to config IDs during tests.
    :param cache: Cache to use.
    :param unparse_with_whitespace: Build test case by adding whitespace between
        nonadjacent tree nodes during unparsing.
    :return: Tuple: (root of the tree, bool whether the tree changed)
    """

    config_ids = [node.id for node in config_nodes]
    config_ids_set = set(config_ids)

    test_builder = PruningTestBuilder(hdd_tree, config_ids_set, with_whitespace=unparse_with_whitespace)
    if cache:
        cache.clear()
        cache.set_test_builder(test_builder)

    test = tester_class(test_builder=test_builder, **tester_config)
    dd = reduce_class(test, cache=cache, id_prefix=id_prefix, **reduce_config)
    c = dd(config_ids)
    if len(c) == 1:
        dd = EmptyDD(test, cache=cache, id_prefix=id_prefix)
        c = dd(c)
    c = set(c)

    def _set_state(node):
        if node.id in config_ids_set:
            node.state = node.KEEP if node.id in c else node.REMOVED
        elif hasattr(node, 'children') and node.state == node.KEEP:
            for child in node.children:
                _set_state(child)
    _set_state(hdd_tree)

    return hdd_tree, len(c) < len(config_ids_set)



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from .ANTLRv4Lexer import ANTLRv4Lexer
from .ANTLRv4Parser import ANTLRv4Parser



#!/usr/bin/env python3

import configparser
import json
import sys


c = configparser.ConfigParser(allow_no_value=True)
with open(sys.argv[1], 'r') as f:
    c.read_file(f)

for s in c.sections():
    for o in c.options(s):
        j = json.loads(c.get(s, o))

c.write(sys.stdout)



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


def coarse_filter(node):
    """
    Config filter to keep nodes with empty replacements only, which is the core
    of the coarse hierarchical delta debugging reduce algorithm.
    """
    return node.replace == ''



# Copyright (c) 2021 Renata Hodovan, Akos Kiss.
# Copyright (c) 2021 Daniel Vince.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import itertools
import logging

from picire import AbstractDD, Outcome

logger = logging.getLogger(__name__)


class HoistingTestBuilder(object):

    def __init__(self, tree, *, with_whitespace=True):
        """
        Initialize the test builder.

        :param tree: Tree representing the current test case.
        :param with_whitespace: Unparse by adding whitespace between nonadjacent
            nodes.
        """
        self.tree = tree
        self.with_whitespace = with_whitespace

    def __call__(self, mapping_config):
        """
        :param mapping_config: A list of mappings of initial configuration
            elements to new ones.
        :return: The unparsed test case with the mappings applied.
        """
        def map(node):
            return mapping.get(node, node)

        mapping = dict(mapping_config)
        return self.tree.unparse(with_whitespace=self.with_whitespace, transform=map)


class MappingMin(AbstractDD):

    def __init__(self, test, *, cache=None, id_prefix=()):
        """
        :param test: A callable tester object.
        :param cache: Cache object to use.
        :param id_prefix: Tuple to prepend to config IDs during tests.
        """

        super().__init__(test=test, split=None, cache=cache, id_prefix=id_prefix)

    def __call__(self, config):
        """
        Compute a mapping of the initial configuration to another (usually
        smaller) but still failing configuration.

        :param config: The initial configuration that will be reduced.
        :return: A mapping of initial configuration elements to new ones.
        """

        def collect_hoistables(node):
            def _collect_hoistables(desc):
                if desc.name == node.name:
                    hoistables.append(desc)
                    return
                if hasattr(desc, 'children') and desc.state == desc.KEEP:
                    for child in desc.children:
                        _collect_hoistables(child)

            hoistables = []
            if hasattr(node, 'children') and node.state == node.KEEP and node.name:
                for child in node.children:
                    _collect_hoistables(child)
            return hoistables

        mapping = {}

        for run in itertools.count():
            logger.info('Run #%d', run)
            logger.info('\tMapping size: %d', len(mapping))
            logger.debug('\tMapping: %r', {c.id: m.id for c, m in mapping.items()})

            for i, (c, m) in enumerate((c, m) for c in config for m in collect_hoistables(mapping.get(c, c))):
                new_mapping = mapping.copy()
                new_mapping[c] = m
                mapping_config = list(new_mapping.items())
                config_id = ('r%d' % run, 'm%d' % i)

                outcome = self._lookup_cache(mapping_config, config_id) or self._test_config(mapping_config, config_id)

                if outcome == Outcome.FAIL:
                    mapping = new_mapping
                    logger.info('\tHoisted')
                    break
            else:
                break

        logger.info('\tDone')
        return mapping


def hoist(hdd_tree, config_nodes, *,
          reduce_class=None, reduce_config=None, tester_class, tester_config,
          id_prefix, cache, unparse_with_whitespace):
    """
    Try hoisting subtrees.

    :param hdd_tree: The root of the tree that the reduce will work with.
    :param config_nodes: Nodes from one level collected by the HDD algorithm.
    :param reduce_class: Unused, present for being compatible with 'prune'
        transformation.
    :param reduce_config: Unused, present for being compatible with 'prune'
        transformation.
    :param tester_class: Reference to a callable class that can decide about the
        interestingness of a test case.
    :param tester_config: Dictionary containing the parameters of the tester
        class init function (except test_builder).
    :param id_prefix: Tuple to prepend to config IDs during tests.
    :param cache: Cache to use.
    :param unparse_with_whitespace: Build test case by adding whitespace between
        nonadjacent tree nodes during unparsing.
    :return: The reduced tree and a boolean value that shows whether the tree
        has changed during hoisting.
    """

    if not config_nodes:
        return hdd_tree, False

    test_builder = HoistingTestBuilder(hdd_tree, with_whitespace=unparse_with_whitespace)
    cache.set_test_builder(test_builder)

    test = tester_class(test_builder=test_builder, **tester_config)
    mapping_min = MappingMin(test, cache=cache, id_prefix=id_prefix)
    mapping = mapping_min(config_nodes)

    def _apply_mapping(node):
        if node in mapping:
            node = mapping[node]
        if hasattr(node, 'children'):
            for i, child in enumerate(node.children):
                node.children[i].replace_with(_apply_mapping(child))
        return node
    hdd_tree = _apply_mapping(hdd_tree)

    if cache:
        cache.clear()

    return hdd_tree, bool(mapping)
